/*
 * Copyright (c) 2004-2005 by John Schmerge.
 * All rights reserved.
 *
 * cron.c
 */

#ifdef __GNUC__
#ident "$Name: micron_0_0_4 $"
#endif

#if (HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <syslog.h>

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "libutil/util.h"
#include "cronpaths.h"
#include "cronfile.h"
#include "job.h"
#include "cron.h"

char * progName = NULL;
int daemonModeFlag = 1;

struct Crontab * tabList = NULL;
struct Crontab systemTab;

/* ********************************************************************
 * sigchld_handler -- handle a child signal
 */
void sigchld_handler(int s)
{
	pid_t pid;
	int status;
	int numberOfChildren = 0;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
	{
		if (WIFEXITED(status))
		{
			syslog(LOG_DEBUG, "Child %d exited with error code %d",
			        pid, WEXITSTATUS(status));

		} else if (WIFSIGNALED(status))
		{
			syslog(LOG_DEBUG, "Child %d was terminated with a signal %d",
			       pid, WTERMSIG(status));
		}

		++numberOfChildren;
	}

	if (numberOfChildren == 0)
	{
		if (pid == -1)
			syslog(LOG_ERR, "Waitpid failed: %m");
	}
}

/* ********************************************************************
 */
void sigterm_handler(int s)
{
	syslog(LOG_INFO, "Recieved SIGTERM. Shutting down.");
	unlink(LOCK_FILE);

	/* This code is conditionally compiled b/c it _really_ isn't
	 * needed unless you're trying to find memory leaks...
	 * The exit call below makes it superfluous.
	 */
#if ( MEM_DEBUG )
	destroyTabList(&tabList);
	destroyJobList(&systemTab.jobList);
#endif

	exit(0);
}

/* ********************************************************************
 */
static void lock_daemon(void)
{
	struct flock lock;
	int fd = -1, n = 0, locked = 0;
	char buffer[11];
	pid_t pid;

	sprintf(buffer, "%d\n", getpid());

	if ((fd = open(LOCK_FILE, O_RDWR | O_CREAT)) >= 0)
	{
		lock.l_start = 0;
		lock.l_len = 0;
		lock.l_pid = getpid();
		lock.l_type = F_WRLCK;
		lock.l_whence = SEEK_SET;

		while ( ! locked )
		{
			if (fcntl(fd, F_SETLKW, &lock) == 0)
			{
				locked = 1;

				if ((n = read(fd, buffer, 11)) > 0)
				{
					buffer[n - 1] = '\0';
					pid = strtoul(buffer, NULL, 10);

					if (kill(pid, 0) != 0)
					{
						fprintf(stderr, "Stale lockfile removed\n");
						ftruncate(fd, 0);
						lseek(fd, 0, SEEK_SET);

						write(fd, buffer, strlen(buffer));

						lock.l_start = 0;
						lock.l_len = 0;
						lock.l_pid = getpid();
						lock.l_type = F_UNLCK;
						lock.l_whence = SEEK_SET;

						fcntl(fd, F_SETLK, &lock);
					} else
					{
						fprintf(stderr,
						        "Found another instance of %s running\n",
						        progName);
						exit(1);
					}
				} else
				{
					write(fd, buffer, strlen(buffer));

					lock.l_start = 0;
					lock.l_len = 0;
					lock.l_pid = getpid();
					lock.l_type = F_UNLCK;
					lock.l_whence = SEEK_SET;

					fcntl(fd, F_SETLK, &lock);
				}
			}
		}

		close(fd);
	} else
	{
		fprintf(stderr, "%s: Could not open %s: %s, exiting\n",
		        progName, LOCK_FILE, strerror(errno));
		exit(1);
	}
}

/* ********************************************************************
 * findJobsToRun -- go through list of cron jobs in a specific crontab,
 *                  and queue any that match the time set for us to run
 */
struct JobQueue * findJobsToRun(struct JobQueue * jq,
                                struct Crontab * tab,
                                time_t * currentTime)
{
	struct JobQueue * jobQueue = jq;
	struct Job * iter = NULL;

	for (iter = tab->jobList; iter != NULL; iter = iter->next)
	{
		if (jobTimeMatch(&(iter->jobTime), currentTime))
			queueJob(&jobQueue, &(tab->tabOptions), iter);
	}

	return jobQueue;
}

/* ********************************************************************
 * mainLoop -- enter an infinite loop where we sleep, wakeup, runjobs,
 *             read crontabs and go back to sleep
 */
void mainLoop(void)
{
	time_t wakeUpTime = 0;
	struct tm * convertedTime = NULL;
	struct Crontab * iterator = NULL;
	struct JobQueue * jobQueue = NULL;

#ifndef SIMULATION
	time_t actualTime = 0, timeLeft = 0, currentTime = 0;
#endif

	/* systemTab is handled a bit differently... we initialize it here */
	systemTab.tabOptions.uid = 0;
	systemTab.lastModificationTime = 0;
	systemTab.jobList = NULL;

	/* figure out how much time we must wait until the start of the next
	 * minute... this way we can start our stuff on a minute boundary
	 */
	(void) time(&wakeUpTime);
	convertedTime = localtime(&wakeUpTime);
	wakeUpTime += (60 - convertedTime->tm_sec);

	while (1)
	{
#ifdef SIMULATION
		while (sleep(1) != 0) { /* empty */ }
#else
		/* sleep until the begining of the next minute */
		while (actualTime < wakeUpTime)
		{
			/* check out config to make sure it hasn't changed */
			checkCrontabs(&systemTab, &tabList);

			/* figure out how much time we have left to sleep */
			(void) time(&currentTime);

			timeLeft = (wakeUpTime < currentTime) ?
			             0 : (wakeUpTime - currentTime);

			actualTime = wakeUpTime - sleep(timeLeft);
		}
#endif
		/*
		 * XXX -- We should revisit this... we can probably do this better...
		 *        it would probably be fairly trivial to calculate how long we
		 *        need to sleep before the next job needs to be run fairly
		 *        painlessly. We'll need to figure out how to wake up however
		 */

		/* re-read crontabs just in case */
		checkCrontabs(&systemTab, &tabList);

		systemTab.next = tabList;

		/* run any jobs that we need to run */
		for (iterator = &systemTab; iterator != NULL; iterator = iterator->next)
		{
			jobQueue = findJobsToRun(jobQueue, iterator, &wakeUpTime);
		}

#ifdef SIMULATION
		if (jobQueue != NULL)
			printf("\n%s", ctime(&wakeUpTime));
#endif
		runJobQueue(&jobQueue);

		wakeUpTime += 60;
	}
}

/* ********************************************************************
 * doInitialization -- do a bunch of initialization stuff
 */
void doInitialization(void)
{
	pid_t pid;
	int fd = 0;
	struct stat st;
	struct sigaction theAction;

	/* we need to pick a directory to run from... the system crontab
	 * directory is as good as any, so we run from there, it makes
	 * several issues involving pathnames easier as well
	 */
	if (stat(CRONTAB_DIR, &st) == 0)
	{
		if (chdir(CRONTAB_DIR) != 0)
		{
			fprintf(stderr, "Error: chdir to %s failed: %s\n",
			        CRONTAB_DIR, strerror(errno));
			exit(1);
		}
	} else
	{
		/* we should probably try to create the cron dir here instead
		 * of exiting
		 */
		fprintf(stderr, "Error: could not stat the dir %s: %s\n",
		        CRONTAB_DIR, strerror(errno));
		exit(1);
	}

	/* fork ourselves into the background */
	if (daemonModeFlag)
	{
		if ((pid = fork()) < 0)
		{
			fprintf(stderr, "Error: could not fork: %s\n", strerror(errno));
			exit(1);

		} else if (pid > 0)
		{
			_exit(0);
		}
	}

	lock_daemon();

	if (daemonModeFlag)
	{
		/* open syslog */
		openlog(progName, LOG_PID, LOG_CRON);

		/* set sid */
		(void) setsid();

		/* close fd's */
		if ((fd = open("/dev/null", O_RDWR)) >= 0)
		{
			dup2(fd, STDIN_FILENO);
			dup2(fd, STDOUT_FILENO);
			dup2(fd, STDERR_FILENO);

			if (fd > STDERR_FILENO)
				close(fd);
		}

	} else
	{
		/* open syslog */
		openlog(progName, LOG_PID | LOG_PERROR, LOG_CRON);
	}

	syslog(LOG_INFO, "%s started", progName);

	/* install signal handlers */
	memset(&theAction, 0, sizeof(struct sigaction));
	theAction.sa_handler = sigchld_handler;
	sigaction(SIGCHLD, &theAction, NULL);

	memset(&theAction, 0, sizeof(struct sigaction));
	theAction.sa_handler = sigterm_handler;

	sigaction(SIGTERM, &theAction, NULL);
	sigaction(SIGINT, &theAction, NULL);
}

/* ********************************************************************
 * usage -- print a usage message
 */
static void usage(void)
{
	printf("Usage: %s [-d] [-v]\n", progName);
}

/* ********************************************************************
 * parseOptions -- parse command line and set global flags based on
 *                 what was parsed
 */
void parseOptions(int argc, char ** argv)
{
	int c = 0;
	char * tmp = NULL;

	progName = ((tmp = strrchr(argv[0], '/')) == NULL ? argv[0] : (tmp + 1));

	while ((c = getopt(argc, argv, "dv")) != -1)
	{
		switch (c)
		{
		 case 'd':
			/* -d means don't run in daemon mode */
			daemonModeFlag = 0;
			break;
		 case 'v':
			puts(PACKAGE_STRING);
			exit(0);
			break;

		 default:
			usage();
			exit(1);
			break;
		}
	}
}

/* ********************************************************************
 * main -- program entry point
 */
int main(int argc, char ** argv)
{
	parseOptions(argc, argv);

	if (getuid())
	{
		fprintf(stderr, "You must be root\n");
		exit(1);
	}

	doInitialization();

	/* enter infinite loop */
	mainLoop();

	return 0;
}

/*********************************************************************
 * $Log: cron.c,v $
 * Revision 1.14  2005/04/25 03:05:41  john
 * Whitespace changes
 *
 * Revision 1.13  2005/04/15 17:22:54  john
 * Size reduction changes and added -v option to program
 *
 * Revision 1.12  2005/04/12 18:09:16  john
 * Changed signal handling to use sigaction() instead of signal()
 *
 * Revision 1.11  2005/04/11 20:01:06  john
 * Added log message when daemon recieves a SIGTERM
 *
 * Revision 1.10  2005/04/03 03:10:36  john
 * Changed one log message to be the same as a log message in another .c
 * file to eliminate the string and got rid of the leading '/' on progName
 *
 * Revision 1.9  2005/03/27 01:18:42  john
 * updated copyright
 *
 * Revision 1.8  2005/03/26 01:42:31  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.7  2005/03/26 00:21:00  john
 * Size reduction changes
 *
 * Revision 1.6.2.2  2005/03/25 13:50:07  john
 * Changed the FREE_ALL_MEM macro to be DEBUG_MEM... Really the only
 * reason that you would want FREE_ALL_MEM is if DEBUG_MEM is enabled.
 *
 * Revision 1.6.2.1  2005/03/25 02:27:03  john
 * #included config.h and fixed up header code
 *
 * Revision 1.6  2003/07/09 04:22:24  john
 * Fix for spelling issue courtesy of Frank Geerlings
 *
 * Revision 1.5  2003/02/11 09:16:27  john
 * size-reducing changes; also removed the job input spec
 *
 * Revision 1.4  2003/02/09 20:46:38  john
 * added job output support; everything but mailing of job output works
 * correctly
 *
 * Revision 1.3  2003/01/22 07:39:48  john
 * Removed typedefs in favor of explicit 'struct' decl's, fixed much
 * brokeness in sigchld_handler, and hacked in necessary changes to
 * accomodate the moving of uid var in the Crontab structure
 *
 * Revision 1.2  2002/10/06 07:18:28  john
 * removed the NOT_CRAZY macro surrounding dead code
 *
 * Revision 1.1.1.1  2002/10/06 06:47:18  john
 * initial import of sources
 *
 */
