/*
 * Copyright (c) 2004-2005 by John Schmerge.
 * All rights reserved.
 *
 * cronparse.c
 */

#if (HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <pwd.h>
#include <syslog.h>

#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "libutil/util.h"
#include "job.h"
#include "cronfile.h"
#include "cronparse.h"

struct KeywordValue
{
	const char * string;
	const int value;
};

static const struct KeywordValue syslogFacilities[] = {
	{ "auth",     LOG_AUTH     }, { "authpriv", LOG_AUTHPRIV },
	{ "cron",     LOG_CRON     }, { "daemon",   LOG_DAEMON   },
	{ "ftp",      LOG_FTP      }, { "kern",     LOG_KERN     },
	{ "local0",   LOG_LOCAL0   }, { "local1",   LOG_LOCAL1   },
	{ "local2",   LOG_LOCAL2   }, { "local3",   LOG_LOCAL3   },
	{ "local4",   LOG_LOCAL4   }, { "local5",   LOG_LOCAL5   },
	{ "local6",   LOG_LOCAL6   }, { "local7",   LOG_LOCAL7   },
	{ "lpr",      LOG_LPR      }, { "mail",     LOG_MAIL     },
	{ "news",     LOG_NEWS     }, { "syslog",   LOG_SYSLOG   },
	{ "user",     LOG_USER     }, { "uucp",     LOG_UUCP     },
	{ NULL,       0            }
};

static const struct KeywordValue syslogPriorities[] = {
	{ "emerg",   LOG_EMERG   }, { "alert",   LOG_ALERT   },
	{ "crit",    LOG_CRIT    }, { "err",     LOG_ERR     },
	{ "warning", LOG_WARNING }, { "notice",  LOG_NOTICE  },
	{ "info",    LOG_INFO    }, { "debug",   LOG_DEBUG   },
	{ NULL,      0           }
};

static const struct KeywordValue logTypes[] = {
	{ "null",        LogTypeDevNull    }, { "filetrunc",   LogTypeFileTrunc  },
	{ "fileappend",  LogTypeFileAppend }, { "syslog",      LogTypeSyslog     },
	{ "mail",        LogTypeMail       }, { "tty",         LogTypeTty        },
	{ NULL,          0                 }
};

/* local prototypes */
static unsigned char * skipLeadingWhitespace(unsigned char * line);

static int isCommentLine(unsigned char * line);

static enum ParseStatus
processLine(unsigned char * line, struct TabOptions * options, struct Job * job);

static enum ParseStatus
processTimeSpec(char * spec, struct JobTime * jt,
                int rangeLow, int rangeHigh, SetJobTimeFunc setFunc);

static enum ParseStatus
processTimeListElem(char * elem, int * begin, int * end, int * step);

static enum ParseStatus
processTimeRange(char * range, int * begin, int * end);

static enum ParseStatus
processNumber(char * numString, int * val);

static enum ParseStatus
processCommandSpec(char * token, struct Job * job);

/* ********************************************************************
 */
int parseTab(const char * fname, struct Crontab * tab)
{
	FILE * f = NULL;
	unsigned char * line = NULL;
	int lineno = 1, jobsPresent = 0;
	struct Job job, * jobPtr = NULL;
	enum ParseStatus rc = ParseOk;

	syslog(LOG_INFO, "Parsing crontab file %s", fname);

	if ((f = fopen(fname, "r")) == NULL)
	{
		syslog(LOG_ERR, "Couldn't open file %s: %m", fname);
		return jobsPresent;
	}

	tab->jobList = NULL;

	while ((line = readLine(f)) != NULL)
	{
		memset(&job, 0, sizeof(job));

		if ((rc = processLine(line, &(tab->tabOptions), &job)) != ParseOk)
		{
			syslog(LOG_ERR, "Parse of crontab file %s failed on line %d",
			       fname, lineno);

			/* XXX Should we do something here besides bitch? */

			/* XXX We need to break out of this loop if the return code
			 *     is something involving incorrect 'OPTION' settings...
			 *     If an option is fubar, then chances are that we need
			 *     to abort processing of this crontab.
			 */
		}

		if (job.command != NULL)
		{
			++jobsPresent;

			jobPtr = (struct Job *) safeMalloc(sizeof(struct Job));
			memcpy(jobPtr, &job, sizeof(struct Job));

			jobPtr->next = tab->jobList;
			tab->jobList = jobPtr;
		}

		++lineno;
	}

	fclose(f);

	return jobsPresent;
}

/* ********************************************************************
 */
static unsigned char * skipLeadingWhitespace(unsigned char * line)
{
	while (*line && ((*line == ' ') || (*line == '\t')))
		++line;

	return line;
}

/* ********************************************************************
 */
static int isCommentLine(unsigned char * line)
{
	if (*line == '\n' || *line == '#' || *line == '\0')
		return 1;
	else
		return 0;
}

/* ********************************************************************
 */
static enum ParseStatus
getValue(const char * value, const struct KeywordValue * valueList,
         unsigned char * retVal)
{
	int i = 0;

	for (i = 0; valueList[i].string != NULL; ++i)
	{
		if (strcmp(valueList[i].string, value) == 0)
		{
			*retVal = valueList[i].value;
			return ParseOk;
		}
	}

	syslog(LOG_ERR, "Invalid keyword value: '%s'", value);

	return ParseBadValue;
}

/* ********************************************************************
 */
static enum ParseStatus
processOptionLine(struct TabOptions * options, unsigned char * line)
{
	char * keyword = NULL, * value = NULL, * strtokBuf = NULL;
	enum ParseStatus rc = ParseOk;

	/* XXX -- we really need to clean up this bit... we shouldn't get
	 *        here if we aren't an option line. Also returning ParseOk
	 *        is a little questionable
	 */
	/* bail out if we aren't a comment line */
	if (*line != '#')
		return rc;

	++line;

	/* skip whitespace after the '#' */
	line = skipLeadingWhitespace(line);

	/* see if we're an option spec */
	if (strncmp(line, "OPTION", sizeof("OPTION") - 1) != 0)
		return rc;

	/* forward the line ptr past the 'OPTION' keyword */
	line = (unsigned char *) strstr(line, "OPTION");
	line += (sizeof("OPTION") - 1);

	/* skip whitespace */
	line = skipLeadingWhitespace(line);

	if ((keyword = strtok_r(line, " \t", &strtokBuf)) != NULL)
		value = strtok_r(NULL, " \t", &strtokBuf);

	if ((keyword != NULL) && (value != NULL))
	{
		if (strcmp(keyword, "LogType") == 0)
		{
			rc = getValue(value, logTypes, &(options->loggingType));

		} else if (strcmp(keyword, "SyslogFacility") == 0)
		{
			rc = getValue(value, syslogFacilities, &options->syslogFacility);

		} else if (strcmp(keyword, "SyslogPriority") == 0)
		{
			rc = getValue(value, syslogPriorities, &options->syslogPriority);

		} else if (strcmp(keyword, "LogFileName") == 0)
		{
			safeFree(options->logFileName);
			options->logFileName = safeStrdup(value);
		} else
		{
			syslog(LOG_ERR, "Invalid option keyword: %s", keyword);
			rc = ParseBadKeyword;
		}
	} else
	{
		rc = ParseBadOptionLine;
	}

	return rc;
}

/* ********************************************************************
 */
static enum ParseStatus processLine(unsigned char * line,
                                    struct TabOptions * options,
                                    struct Job * job)
{
	enum ParseStatus rc = ParseOk;
	char * strtokBuf = NULL, * token = NULL;

	/* skip leading whitespace */
	line = skipLeadingWhitespace(line);

	/* XXX -- revisit handling of rc here */
	rc = processOptionLine(options, line);

	if (isCommentLine(line))
		return rc;

	memset(&job->jobTime, 0, sizeof(job->jobTime));

	/* XXX -- What happens when you specify an incomplete job line */
	if (((token = strtok_r(line, " \t", &strtokBuf)) == NULL) ||
	    ((rc = processTimeSpec(token, &job->jobTime, MINUTE_MIN, MINUTE_MAX,
	                           setJobTimeMinute)) != ParseOk))
	{
		return rc;
	}

	if (((token = strtok_r(NULL, " \t", &strtokBuf)) == NULL) ||
	    ((rc = processTimeSpec(token, &job->jobTime, HOUR_MIN, HOUR_MAX,
	                           setJobTimeHour)) != ParseOk))
	{
		return rc;
	}

	if (((token = strtok_r(NULL, " \t", &strtokBuf)) == NULL) ||
	    ((rc = processTimeSpec(token, &job->jobTime, DOM_MIN, DOM_MAX,
	                           setJobTimeDom)) != ParseOk))
	{
		return rc;
	}

	if (((token = strtok_r(NULL, " \t", &strtokBuf)) == NULL) ||
	    ((rc = processTimeSpec(token, &job->jobTime, MONTH_MIN, MONTH_MAX,
	                           setJobTimeMonth)) != ParseOk))
	{
		return rc;
	}

	if (((token = strtok_r(NULL, " \t", &strtokBuf)) == NULL) ||
	    ((rc = processTimeSpec(token, &job->jobTime, DOW_MIN, DOW_MAX,
	                           setJobTimeDow)) != ParseOk))
	{
		return rc;
	}

	if (((token = strtok_r(NULL, "\n", &strtokBuf)) == NULL) ||
	    ((rc = processCommandSpec(token, job)) != ParseOk))
	{
		return rc;
	}

	return rc;
}

/* ********************************************************************
 */
static enum ParseStatus processTimeSpec(char * spec, struct JobTime * jt,
                                        int rangeLow, int rangeHigh,
                                        SetJobTimeFunc setFunc)
{
	enum ParseStatus rc = ParseOk;
	char * strtokBuf = NULL, * tok = NULL, * buf = NULL;
	int begin = 0, end = -1, step = 1, i = 0;

	buf = spec;

	while ((tok = strtok_r(buf, ",", &strtokBuf)) != NULL)
	{
		/* this statement makes this loop a little bit more regular, rather
		 * than do the special first case of the strtok above
		 */
		buf = NULL;

		/* set these to the initial defaults... */
		begin = -1;
		end = -1;
		step = 1;

		if ((rc = processTimeListElem(tok, &begin, &end, &step)) != ParseOk)
		{
			return rc;
		}

		/* this is a special case... */
		if ((begin == -1) && (end == -1))
		{
			begin = rangeLow;
			end = rangeHigh;
		}

		/* do a little bit of semantic processing...
		 * check to make sure our range spec is within the min & max
		 */
		if ((begin > end) || (begin < rangeLow) || (end > rangeHigh))
			return ParseInvalidRange;

		/* make sure that step is positive & not greater than the
		 * range being stepped over
		 */
		if ((step < 1) || ( (step > 1) && (step >= (end - begin)) ) )
			return ParseInvalidStep;

		/* check to make sure a bonehead didn't do a single number followed by
		 * a step... we don't want to get too confused
		 *
		 * XXX should this instead mean <number>-MAX/<step> ?
		 */
		if ((step > 1) && (begin == end))
			return ParseStepNoEffect;

		i = begin;
		do {
			setFunc(jt, i);
			i += step;
		} while (i <= end);
	}

	return ParseOk;
}

/* ********************************************************************
 */
static enum ParseStatus processTimeListElem(char * elem, int * begin,
                                            int * end, int * step)
{
	enum ParseStatus rc = ParseOk;
	char * rangeString = NULL, * stepString = NULL;

	rangeString = elem;

	if ((stepString = strchr(elem, '/')) != NULL)
	{
		*stepString = '\0';
		++stepString;
	}

	if (stepString != NULL)
	{
		if ((rc = processNumber(stepString, step)) != ParseOk)
			return rc;
	}

	if ((rc = processTimeRange(rangeString, begin, end)) != ParseOk)
		return rc;

	return ParseOk;
}

/* ********************************************************************
 */
static enum ParseStatus processTimeRange(char * range, int * begin, int * end)
{
	enum ParseStatus rc = ParseOk;
	char * low = NULL, * high = NULL;

	if ((*range == '*') && (range[1] == '\0'))
	{
		*begin = -1;
		*end = -1;
	} else
	{
		low = range;

		if ((high = strchr(range, '-')) != NULL)
		{
			*high = '\0';
			++high;
		}

		if ((rc = processNumber(low, begin)) != ParseOk)
			return rc;

		if (high != NULL)
		{
			if ((rc = processNumber(high, end)) != ParseOk)
				return rc;
		} else
		{
			*end = *begin;
		}
	}

	return ParseOk;
}

/* ********************************************************************
 */
static enum ParseStatus processNumber(char * numString, int * val)
{
	enum ParseStatus rc = ParseOk;
	long int n = 0;
	char * endptr = NULL;

	if ( (*numString == '\0') ||
	     ( (n = strtol(numString, &endptr, 10) ) == LONG_MIN) ||
	     (n == LONG_MAX) ||
	     (*endptr != '\0'))
	{
		syslog(LOG_ERR, "Invalid number: %s", numString);
		rc = ParseInvalidNumber;
	} else
	{
		*val = n;
	}

	return rc;
}

/* ********************************************************************
 */
static enum ParseStatus processCommandSpec(char * token, struct Job * job)
{
	enum ParseStatus rc = ParseOk;

	/* make sure there's actually a command to execute */
	token = skipLeadingWhitespace(token);

	if (strlen(token) != 0)
		job->command = safeStrdup(token);
	else
		rc = ParseEmptyCommand;

	return rc;
}

/* ********************************************************************
 * $Log: cronparse.c,v $
 * Revision 1.12  2005/04/25 03:29:07  john
 * Modified the CronFile struct to contained unsigned char's for enum values
 * instead of int's and added support for specifying logfile name.
 *
 * Revision 1.11  2005/04/22 02:28:19  john
 * Reworked the option parsing code to be a little cleaner as well as to
 * actually return meaningful errors
 *
 * Revision 1.10  2005/04/15 17:23:39  john
 * Size reduction changes
 *
 * Revision 1.9  2005/03/27 01:19:27  john
 * updated copyright
 *
 * Revision 1.8  2005/03/26 01:42:31  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.7  2005/03/26 00:23:55  john
 * Size reduction changes
 *
 * Revision 1.6.2.1  2005/03/25 02:27:03  john
 * #included config.h and fixed up header code
 *
 * Revision 1.6  2004/02/14 02:46:28  john
 * Clean up of the parse routine return codes
 *
 * Revision 1.5  2003/05/12 04:04:21  john
 * Abstracted whitespace skipping code into a function
 *
 * Revision 1.4  2003/02/11 09:16:27  john
 * size-reducing changes; also removed the job input spec
 *
 * Revision 1.3  2003/02/09 20:46:38  john
 * added job output support; everything but mailing of job output works
 * correctly
 *
 * Revision 1.2  2003/01/22 07:48:03  john
 * Removed typedefs and added a majority of option parsing code,
 *
 * Revision 1.1.1.1  2002/10/06 06:47:18  john
 * initial import of sources
 *
 */
