/*
 * Copyright (c) 2004-2005 by John Schmerge.
 * All rights reserved.
 */

#ifdef __GNUC__
#ident "$Name: micron_0_0_4 $"
#endif

#if (HAVE_CONFIG_H)
# include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <paths.h>

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "libutil/util.h"
#include "cronpaths.h"

/* Type Definitions */
enum Action
{
	ACTION_UNKNOWN,
	ACTION_EDIT,
	ACTION_REMOVE,
	ACTION_LIST,
	ACTION_INSERT
};

/* Globals */
char * progName = NULL;
enum Action theAction = ACTION_UNKNOWN;
const char * requestedUser = NULL;
const char * replacementFilename = NULL;

/* ********************************************************************
 */
void usage(void)
{
	printf("usage:\t%s [-u user] file\n"
           "\t%s [-u user] { -e | -l | -r }\n"
           "\t\t(default operation is replace, per 1003.2)\n"
           "\t-e\t(edit user's crontab)\n"
           "\t-l\t(list user's crontab)\n"
           "\t-r\t(delete user's crontab)\n"
           "\t-v\t(print version)\n" , progName, progName);
}

/* ********************************************************************
 */
void parseOptions(int argc, char *** argv)
{
	int c = 0;
	char * tmp = NULL, ** argvLocal = *argv;

	/* strip the leading path off of the program name */
	progName = (tmp = strrchr((*argv)[0], '/')) ? tmp + 1 : (*argv)[0];
	(*argv)[0] = progName;

	while ((c = getopt(argc, argvLocal, "herlu:v")) != -1)
	{
		switch (c)
		{
		 case 'e':
			if (theAction != ACTION_UNKNOWN)
			{
				fprintf(stderr, "%s: only one action may be specified\n",
				        progName);
				usage();
				exit(1);
			}

			theAction = ACTION_EDIT;
			break;

		 case 'r':
			if (theAction != ACTION_UNKNOWN)
			{
				fprintf(stderr, "%s: only one action may be specified\n",
				        progName);
				usage();
				exit(1);
			}

			theAction = ACTION_REMOVE;
			break;

		 case 'l':
			if (theAction != ACTION_UNKNOWN)
			{
				fprintf(stderr, "%s: only one action may be specified\n",
				        progName);
				usage();
				exit(1);
			}

			theAction = ACTION_LIST;
			break;

		 case 'u':
			requestedUser = optarg;
			break;

		 case 'v':
			printf("%s\n", PACKAGE_STRING);
			exit(0);
			break;

		 case 'h':
			usage();
			exit(0);
			break;

		 default:
			usage();
			exit(1);
			break;
		}
	}

	/* By default, we insert the file specified into the crontab dir */
	if (theAction == ACTION_UNKNOWN)
		theAction = ACTION_INSERT;

	/* Check to make sure that we were given a file name on the
	 * command-line if the action is insertion
	 */
	if (theAction == ACTION_INSERT)
	{
		if (argvLocal[optind] != NULL)
		{
			replacementFilename = argvLocal[optind];
			++optind;
		} else
		{
			fprintf(stderr, "%s: File name must be specified for replacement "
			                "function\n", progName);
			usage();
			exit(1);
		}
	}

	/* Print warning about excess arguments being ignored */
	if (argvLocal[optind] != NULL)
		fprintf(stderr, "%s: Warning: excess arguments specified supplied on "
		                "command line have been ignored\n", progName);
}

/* ********************************************************************
 */
char * createCronFilePath(const char *userName)
{
	char * rv = NULL;
	int len = 0;

	/* Figure out how long our resulting path will be...
	 * sizeof(CRONTAB_DIR) includes the trailing NUL, strlen(userName)
	 * is the length of the user name and the '+ 1' is for the directory
	 * seperator
	 */
	len = sizeof(CRONTAB_DIR) + strlen(userName) + 1;

	rv = (char *) safeMalloc(len);

	strncpy(rv, CRONTAB_DIR, len);
	strncat(rv, "/", len);
	strncat(rv, userName, len);

	return rv;
}

/* ********************************************************************
 */
int copyFile(FILE * dest, FILE * src)
{
	int rc = 0, c = 0;

	while ((c = getc(src)) != EOF)
	{
		putc(c, dest);
		++rc;
	}

	return rc;
}

/* ********************************************************************
 */
const char * getEditor(void)
{
	const char * editor = NULL;

	if ((editor = getenv("EDITOR")) != NULL)
		return editor;
	else if ((editor = getenv("VISUAL")) != NULL)
		return editor;
	else
		return DEFAULT_EDITOR;
}

/* ********************************************************************
 */
int doEditAction(const struct passwd * pwent)
{
	int rc = 0, editFilePathLen = 0, editFileFd = -1, waitStatus = 0;
	char * cronFilePath = NULL, * editFilePath = NULL, * tmpDirPath = NULL;
	const char * editorProg;
	pid_t editorPid = -1;

	FILE * editFilePtr = NULL, * cronFilePtr = NULL;

	editorProg = getEditor();
	cronFilePath = createCronFilePath(pwent->pw_name);

	tmpDirPath = getenv("TMPDIR");
	if (tmpDirPath == NULL)
		tmpDirPath = _PATH_TMP;

	/* tmpdir + / + progName + .XXXXXX + NUL */
	editFilePathLen = strlen(tmpDirPath) + 1 + strlen(progName) + 7 + 1;
	editFilePath = (char *) safeMalloc(editFilePathLen);
	strncpy(editFilePath, tmpDirPath, editFilePathLen);
	strncat(editFilePath, "/", editFilePathLen);
	strncat(editFilePath, progName, editFilePathLen);
	strncat(editFilePath, ".XXXXXX", editFilePathLen);

	editFileFd = mkstemp(editFilePath);
	chmod(editFilePath, 0600);
	editFilePtr = fdopen(editFileFd, "w");

	/* make temporary copy of crontab */
	if ((cronFilePtr = fopen(cronFilePath, "r")) != NULL)
	{
		copyFile(editFilePtr, cronFilePtr);
		fclose(cronFilePtr);
	}

	/* Close file in case the editor removes the old copy on save */
	fclose(editFilePtr);

	chown(editFilePath, pwent->pw_uid, pwent->pw_gid);

	/* Exec editor */
	if ((editorPid = fork()) == 0)
	{
		initgroups(pwent->pw_name, pwent->pw_gid);
		setgid(pwent->pw_gid);
		setuid(pwent->pw_uid);

		execlp(editorProg, editorProg, editFilePath, NULL);

		fprintf(stderr, "Exec of editor %s failed: %s\n", editorProg,
		        strerror(errno));
		_exit(1);
	} else if (editorPid < 0)
	{
		fprintf(stderr, "Fork failed: %s\n", strerror(errno));
		unlink(editFilePath);
		safeFree(cronFilePath);
		safeFree(editFilePath);
		exit(1);
	}

	if (waitpid(editorPid, &waitStatus, 0) != editorPid)
	{
		fprintf(stderr, "Wait failed: %s\n", strerror(errno));
		unlink(editFilePath);
		safeFree(cronFilePath);
		safeFree(editFilePath);
		exit(1);
	}

	if ((editFilePtr = fopen(editFilePath, "r")) == NULL)
	{
		fprintf(stderr, "Couldn't re-open edit file: %s\n", strerror(errno));
		unlink(editFilePath);
		safeFree(cronFilePath);
		safeFree(editFilePath);
		exit(1);
	}

	/* XXX - check crontab syntax */

	/* copy/overwrite current crontab */
	if ((cronFilePtr = fopen(cronFilePath, "w")) != NULL)
	{
		chown(cronFilePath, 0, 0);
		chmod(cronFilePath, 0600);
		copyFile(cronFilePtr, editFilePtr);
		fclose(cronFilePtr);
	} else
	{
		fprintf(stderr, "Couldn't open crontab file: %s\n", strerror(errno));
		unlink(editFilePath);
		safeFree(cronFilePath);
		safeFree(editFilePath);
		exit(1);
	}

	fclose(editFilePtr);
	unlink(editFilePath);

	safeFree(cronFilePath);
	safeFree(editFilePath);

	return rc;
}

/* ********************************************************************
 */
int doRemoveAction(const char * userName)
{
	char * cronFilePath = NULL;
	int rc = 0;

	cronFilePath = createCronFilePath(userName);

	if (unlink(cronFilePath) != 0)
	{
		if (errno == ENOENT)
			fprintf(stderr, "No crontab for %s\n", userName);
		else
			fprintf(stderr, "Error: couldn't remove crontab file for "
			                "user %s:%s\n", userName, strerror(errno));
		rc = 1;
	}

	safeFree(cronFilePath);

	return rc;
}

/* ********************************************************************
 */
int doListAction(const char * userName)
{
	char * cronFilePath = NULL;
	FILE * f = NULL;
	int rc = 0;

	cronFilePath = createCronFilePath(userName);

	if ((f = fopen(cronFilePath, "r")) != NULL)
	{
		copyFile(stdout, f);
		fclose(f);
	} else
	{
		if (errno == ENOENT)
			fprintf(stderr, "No crontab for %s\n", userName);
		else
			fprintf(stderr, "Error: couldn't open file %s: %s\n",
			        cronFilePath, strerror(errno));
		rc = 1;
	}

	safeFree(cronFilePath);

	return rc;
}

/* ********************************************************************
XXX - Need to lock file once we create it
 */
int doReplaceAction(const char * userName, const char * replacementFilePath)
{
	char * cronFilePath = NULL;
	FILE * src = NULL, * dst = NULL;
	int rc = 0;

	cronFilePath = createCronFilePath(userName);

	if ((unlink(cronFilePath) != 0) && (errno != ENOENT))
	{
		fprintf(stderr, "Error: couldn't remove old crontab file for "
		                "user %s:%s\n", userName, strerror(errno));
		rc = 1;
	}

	if ((src = fopen(replacementFilePath, "r")) == NULL)
	{
		fprintf(stderr, "Error: couldn't open %s for reading: %s\n",
		        replacementFilePath, strerror(errno));
		rc = 1;
	} else if ((dst = fopen(cronFilePath, "w")) == NULL)
	{
		fprintf(stderr, "Error: couldn't open %s for writing: %s\n",
		        cronFilePath, strerror(errno));
		rc = 1;
	} else
	{
		chown(cronFilePath, 0, 0);
		chmod(cronFilePath, 0600);
		copyFile(dst, src);
		fclose(dst);
	}

	if (src != NULL)
		fclose(src);

	safeFree(cronFilePath);

	return rc;
}

/* ********************************************************************
 */
int main(int argc, char ** argv)
{
	struct passwd * pwent = NULL;
	uid_t userId = 0, effectiveId = 0;
	int rc = 0;

	parseOptions(argc, &argv);

	/* Figure out which crontab we're supposed to be performing an
	 * action upon ... We do this here so that we can bail early if
	 * things don't smell right
	 */
	userId = getuid();
	effectiveId = geteuid();

	if ((pwent = getpwuid(userId)) == NULL)
	{
		fprintf(stderr, "No passwd entry for uid %d\n", userId);
		exit(1);
	}

	if (requestedUser != NULL)
	{
		if (userId != 0)
		{
			fprintf(stderr, "Only root can modify another user's crontab\n");
			exit(1);
		}

		if ((pwent = getpwnam(requestedUser)) == NULL)
		{
			fprintf(stderr, "Couldn't find passwd entry for user %s",
			        requestedUser);
			exit(1);
		}
	}

	/* If we need to run as a user other than ourselves, de-elevate privledges
	 * and continue
	 */
	switch (theAction)
	{
	 case ACTION_EDIT:
		rc = doEditAction(pwent);
		break;

	 case ACTION_REMOVE:
		rc = doRemoveAction(pwent->pw_name);
		break;

	 case ACTION_LIST:
		rc = doListAction(pwent->pw_name);
		break;

	 case ACTION_INSERT:
		/* XXX - check crontab syntax */
		rc = doReplaceAction(pwent->pw_name, replacementFilename);
		break;

	 default:
		abort();
		break;
	}

	return rc;
}

/* ********************************************************************
 * $Log: crontab.c,v $
 * Revision 1.12  2005/05/04 10:41:53  john
 * Added RCS Name macro to file
 *
 * Revision 1.11  2005/04/15 17:24:26  john
 * size reduction changes
 *
 * Revision 1.10  2005/04/12 19:57:26  john
 * changed how util.h gets included
 *
 * Revision 1.9  2005/04/03 03:11:15  john
 * Added -v option to the program
 *
 * Revision 1.8  2005/04/02 18:43:07  john
 * First working version of the crontab program
 *
 * Revision 1.7  2005/03/26 01:42:31  john
 * Merge from branch micron_autoconf
 *
 * Revision 1.6.2.1  2005/03/25 02:27:03  john
 * #included config.h and fixed up header code
 *
 * Revision 1.6  2005/02/26 13:09:10  john
 * Intermediate checkin.
 *
 * Revision 1.5  2004/02/26 23:38:54  john
 * intermediate check in for colaborative devel
 *
 * Revision 1.4  2003/04/27 23:09:39  john
 * Fixed warning by include stdlib.h
 *
 * Revision 1.3  2003/02/11 09:16:27  john
 * size-reducing changes; also removed the job input spec
 *
 * Revision 1.2  2003/01/22 07:48:51  john
 * removed typedefs in favor of explicit declarations
 *
 * Revision 1.1.1.1  2002/10/06 06:47:18  john
 * initial import of sources
 *
 */
